import React from 'react';
import PropTypes from 'prop-types';
import { connect } from 'react-redux';
import isFunction from 'lodash/isFunction';
import omit from 'lodash/omit';
import isArray from 'lodash/isArray';
import isEqual from 'lodash/isEqual';

let globalModels = {};
let errorHandler = null;

export function setGlobalModels(models) {
  globalModels = Object.keys(models).reduce(
    (globalModels, key) =>
      models[key].modelName === key
        ? { ...globalModels, [key]: models[key] }
        : globalModels,
    {}
  );
}

export function setErrorHandler(value) {
  errorHandler = isFunction(value) && value;
}

class Wrapper extends React.Component {
  componentDidMount() {
    const { _actions, dispatch, _handleError } = this.props;
    _actions.forEach(async ({ action }) => {
      try {
        await dispatch(action);
      } catch (e) {
        _handleError(e, this.props);
      }
    });
  }

  componentDidUpdate({ _actions: actions }) {
    const { _actions: newActions, dispatch, _handleError } = this.props;

    (newActions || [])
      .filter(
        ({ name, modelName, actionName, actionParams }) =>
          actions.findIndex(
            item =>
              item.name === name &&
              item.modelName === modelName &&
              item.actionName === actionName &&
              isEqual(item.actionParams, actionParams)
          ) === -1
      )
      .forEach(async ({ action }) => {
        try {
          await dispatch(action);
        } catch (e) {
          _handleError(e, this.props);
        }
      });
  }

  render() {
    const { Component } = this.props;
    const props = omit(
      this.props,
      'Component',
      '_actions',
      '_handleError',
      'state'
    );
    return <Component {...props} />;
  }
}

Wrapper.propTypes = {
  Component: PropTypes.oneOfType([PropTypes.func, PropTypes.object]),
  _actions: PropTypes.array,
  dispatch: PropTypes.func
};

const emptyComponentProps = () => ({});

export default function withModels(Component, componentProps, options = {}) {
  const models = {
    ...globalModels,
    ...Object.keys(options).reduce(
      (models, key) =>
        options[key].modelName === key
          ? { ...models, [key]: options[key] }
          : models,
      {}
    )
  };

  const { omitErrorsFor } = options;

  function _handleError(e, props, action) {
    let rules = omitErrorsFor;
    if (typeof omitErrorsFor === 'function') {
      rules = omitErrorsFor(props);
    }

    const omitError =
      rules &&
      isArray(rules) &&
      rules.length > 0 &&
      rules.findIndex(omitAction => isEqual(omitAction, action)) === -1;

    (!omitError && errorHandler && errorHandler(e, props)) || console.error(e);
  }

  const ConnectedComponent = connect(
    state => state,
    null,
    (stateProps, dispatchProps, props) => {
      const allProps =
        (componentProps || emptyComponentProps)({
          ...props,
          ...dispatchProps,
          state: stateProps
        }) || {};
      const actionKeys = Object.keys(allProps).filter(
        name => isFunction(allProps[name]) && allProps[name].modelName
      );

      const _actions = actionKeys.map(name => ({
        name,
        action: allProps[name],
        modelName: allProps[name].modelName,
        actionName: allProps[name].actionName,
        actionParams: allProps[name].actionParams
      }));

      const actionsProps = _actions.reduce(
        (props, { name, modelName, actionName, actionParams }) => {
          const selectors = models[modelName](stateProps);
          const data =
            (typeof selectors[`${actionName}Meta`] === 'function' &&
              selectors[`${actionName}Meta`](...actionParams)) ||
            {};
          const result =
            (typeof selectors[actionName] === 'function' &&
              selectors[actionName](...actionParams)) ||
            null;

          return {
            ...props,
            [name]: result,
            [`${name}Requested`]: data.requested === true,
            [`${name}Requesting`]: data.requesting === true,
            [`${name}Error`]: data.error
          };
        },
        {}
      );

      const restProps = omit(allProps, actionKeys);

      return {
        ...dispatchProps,
        ...props,
        ...actionsProps,
        ...restProps,
        _actions,
        _handleError,
        Component
      };
    }
  )(Wrapper);

  const componentName = Component.displayName || Component.name || 'Component';
  ConnectedComponent.displayName = `WithModels(${componentName})`;

  return ConnectedComponent;
}
